"""FileSystem storage backend for Ralph"""

import datetime
import logging
import sys
from pathlib import Path

from ralph.defaults import FS_STORAGE_DEFAULT_PATH
from ralph.utils import now

from ..mixins import HistoryMixin
from .base import BaseStorage

logger = logging.getLogger(__name__)


class FSStorage(HistoryMixin, BaseStorage):
    """FileSystem storage backend"""

    name = "fs"

    def __init__(self, path=FS_STORAGE_DEFAULT_PATH):
        """Create the path directory if it does not exist"""

        self._path = Path(path)
        if not self._path.is_dir():
            logger.info("FS storage directory doesn't exist, creating: %s", self._path)
            self._path.mkdir(parents=True)

        logger.debug("File system storage path: %s", self._path)

    def _get_filepath(self, name, strict=False):
        """Return path of the archive in the FS storage, or throws an exception if not found"""

        file_path = self._path / Path(name)
        if strict and not file_path.exists():
            msg = "%s file does not exist"
            logger.error(msg, file_path)
            raise FileNotFoundError(msg % file_path)
        return file_path

    def _details(self, name):
        """Get name archive details"""

        file_path = self._get_filepath(name)
        stats = file_path.stat()

        return {
            "filename": name,
            "size": stats.st_size,
            "modified_at": datetime.datetime.fromtimestamp(
                int(stats.st_mtime), tz=datetime.timezone.utc
            ).isoformat(),
        }

    def list(self, details=False, new=False):
        """List files in the storage backend"""

        archives = [archive.name for archive in self._path.iterdir()]
        logger.debug("Found %d archives", len(archives))

        if new:
            archives = set(archives) - set(self.get_command_history(self.name, "fetch"))
            logger.debug("New archives: %d", len(archives))

        for archive in archives:
            yield self._details(archive) if details else archive

    def url(self, name):
        """Get `name` file absolute URL"""

        return str(self._get_filepath(name).resolve(strict=True))

    def read(self, name, chunk_size=4096):
        """Read `name` file and stream its content by chunks of a given size"""

        logger.debug("Getting archive: %s", name)

        with self._get_filepath(name).open("rb") as file:
            while chunk := file.read(chunk_size):
                sys.stdout.buffer.write(chunk)

        details = self._details(name)
        # Archive is supposed to have been fully fetched, add a new entry to
        # the history.
        self.append_to_history(
            {
                "backend": self.name,
                "command": "fetch",
                "id": name,
                "filename": details.get("filename"),
                "size": details.get("size"),
                "fetched_at": now(),
            }
        )

    def write(self, name, chunk_size=4096, overwrite=False):
        """Write content to the `name` target"""

        logger.debug("Creating archive: %s", name)

        file_path = self._get_filepath(name)
        if file_path.is_file() and not overwrite:
            msg = "%s already exists and overwrite is not allowed"
            logger.error(msg, name)
            raise FileExistsError(msg, name)

        with file_path.open("w") as file:
            while chunk := sys.stdin.read(chunk_size):
                file.write(chunk)

        details = self._details(name)
        # Archive is supposed to have been fully created, add a new entry to
        # the history.
        self.append_to_history(
            {
                "backend": self.name,
                "command": "push",
                "id": name,
                "filename": details.get("filename"),
                "size": details.get("size"),
                "pushed_at": now(),
            }
        )



"""Base storage backend for Ralph"""

from abc import ABC, abstractmethod


class BaseStorage(ABC):
    """Base storage backend interface"""

    name = "base"

    @abstractmethod
    def list(self, details=False, new=False):
        """List files in the storage backend"""

    @abstractmethod
    def url(self, name):
        """Get `name` file absolute URL"""

    @abstractmethod
    def read(self, name, chunk_size=4096):
        """Read `name` file and stream its content by chunks of a given size"""

    @abstractmethod
    def write(self, name, chunk_size=4096, overwrite=False):
        """Write content to the `name` target"""



"""Ralph CLI entrypoint"""

import json
import logging
import re
import sys
from inspect import signature

import click
from click_option_group import optgroup

from ralph.backends import BackendTypes
from ralph.defaults import (
    DEFAULT_BACKEND_CHUNCK_SIZE,
    DEFAULT_GELF_PARSER_CHUNCK_SIZE,
    ENVVAR_PREFIX,
    DatabaseBackends,
    Parsers,
    StorageBackends,
)
from ralph.exceptions import UnsupportedBackendException
from ralph.logger import configure_logging
from ralph.utils import (
    get_backend_type,
    get_class_from_name,
    get_class_names,
    get_instance_from_class,
    get_root_logger,
)

# cli module logger
logger = logging.getLogger(__name__)

# Lazy evaluations
DATABASE_BACKENDS = (lambda: [backend.value for backend in DatabaseBackends])()
PARSERS = (lambda: [parser.value for parser in Parsers])()
STORAGE_BACKENDS = (lambda: [backend.value for backend in StorageBackends])()
BACKENDS = (lambda: DATABASE_BACKENDS + STORAGE_BACKENDS)()


class CommaSeparatedKeyValueParamType(click.ParamType):
    """Comma separated key=value parameter type"""

    name = "key=value,key=value"

    def convert(self, value, param, ctx):
        """Split value by comma and equal sign to return a dict build with key/value pairs"""

        # Parse options string
        try:
            options = dict(option.split("=") for option in value.split(","))
        except ValueError:
            self.fail(
                "You should provide key=value pairs separated by commas, e.g. foo=bar,bar=2",
                param,
                ctx,
            )

        # Cast simple types
        for key, value_ in options.items():
            if value_ == "":
                options.update({key: None})
            elif re.match("^true$", value_, re.IGNORECASE):
                options.update({key: True})
            elif re.match("^false$", value_, re.IGNORECASE):
                options.update({key: False})
            elif re.match(r"^[0-9]+\.[0-9]+$", value_):
                options.update({key: float(value_)})
            elif re.match("^[0-9]+$", value_):
                options.update({key: int(value_)})

        return options


@click.group(name="ralph")
@click.option(
    "-v",
    "--verbosity",
    type=click.Choice(["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]),
    metavar="LVL",
    required=False,
    help="Either CRITICAL, ERROR, WARNING, INFO (default) or DEBUG",
)
def cli(verbosity=None):
    """Ralph is a stream-based tool to play with your logs"""

    configure_logging()
    if verbosity is not None:
        level = getattr(logging, verbosity, None)
        get_root_logger().setLevel(level)
        for handler in get_root_logger().handlers:
            handler.setLevel(level)


def backends_options(name=None, backends=None):
    """Backend-related options decorator for Ralph commands"""

    backend_names = get_class_names(backends)

    def wrapper(command):
        command = (
            click.option(
                "-b",
                "--backend",
                type=click.Choice(backend_names),
                required=True,
                help="Backend",
            )
        )(command)

        for backend_name in backend_names:
            backend_class = get_class_from_name(backend_name, backends)

            for parameter in signature(backend_class.__init__).parameters.values():
                if parameter.name == "self":
                    continue
                option = f"--{backend_class.name}-{parameter.name}".replace("_", "-")
                envvar = (
                    f"{ENVVAR_PREFIX}_{backend_class.name}_{parameter.name}".upper()
                )
                option_kwargs = {}
                # If the parameter is a boolean, convert it to a flag option
                if parameter.annotation is bool:
                    option = (
                        f"{option}/--no-{backend_class.name}-{parameter.name}".replace(
                            "_", "-"
                        )
                    )
                    option_kwargs["is_flag"] = True
                elif parameter.annotation is dict:
                    option_kwargs["type"] = CommaSeparatedKeyValueParamType()
                command = (
                    optgroup.option(
                        option,
                        envvar=envvar,
                        default=parameter.default,
                        **option_kwargs,
                    )
                )(command)
            command = (optgroup.group(f"{backend_class.name} backend"))(command)

        command = (cli.command(name=name or command.__name__))(command)
        return command

    return wrapper


@cli.command()
@click.option(
    "-p",
    "--parser",
    type=click.Choice(get_class_names(PARSERS)),
    required=True,
    help="Container format parser used to extract events",
)
@click.option(
    "-c",
    "--chunksize",
    type=int,
    default=DEFAULT_GELF_PARSER_CHUNCK_SIZE,
    help="Parse events by chunks of size #",
)
def extract(parser, chunksize):
    """Extract input events from a container format using a dedicated parser"""

    logger.info(
        "Extracting events using the %s parser (chunk size: %d)", parser, chunksize
    )

    parser = get_class_from_name(parser, PARSERS)()

    for event in parser.parse(sys.stdin, chunksize=chunksize):
        click.echo(event)


@click.argument("archive", required=False)
@backends_options(backends=BACKENDS)
@click.option(
    "-c",
    "--chunk-size",
    type=int,
    default=DEFAULT_BACKEND_CHUNCK_SIZE,
    help="Get events by chunks of size #",
)
def fetch(backend, archive, chunk_size, **options):
    """Fetch an archive or records from a configured backend"""

    logger.info(
        "Fetching data from the configured %s backend (archive: %s | chunk size: %s)",
        backend,
        archive,
        chunk_size,
    )
    logger.debug("Backend parameters: %s", options)

    backend_class = get_class_from_name(backend, BACKENDS)
    backend = get_instance_from_class(backend_class, **options)
    backend_type = get_backend_type(backend_class)

    if backend_type == BackendTypes.STORAGE:
        backend.read(archive, chunk_size=chunk_size)
    elif backend_type == BackendTypes.DATABASE:
        backend.get(chunk_size=chunk_size)
    elif backend_type is None:
        msg = "Cannot find an implemented backend type for backend %s"
        logger.error(msg, backend)
        raise UnsupportedBackendException(msg, backend)


@click.argument("archive", required=False)
@backends_options(backends=BACKENDS)
@click.option(
    "-c",
    "--chunk-size",
    type=int,
    default=DEFAULT_BACKEND_CHUNCK_SIZE,
    help="Get events by chunks of size #",
)
@click.option(
    "-f",
    "--force",
    default=False,
    is_flag=True,
    help="Overwrite existing archives or records",
)
@click.option(
    "-I",
    "--ignore-errors",
    default=False,
    is_flag=True,
    help="Continue writing regardless of raised errors",
)
def push(backend, archive, chunk_size, force, ignore_errors, **options):
    """Push an archive to a configured backend"""

    logger.info("Pushing archive %s to the configured %s backend", archive, backend)
    logger.debug("Backend parameters: %s", options)

    backend_class = get_class_from_name(backend, BACKENDS)
    backend = get_instance_from_class(backend_class, **options)
    backend_type = get_backend_type(backend_class)

    if backend_type == BackendTypes.STORAGE:
        backend.write(archive, overwrite=force)
    elif backend_type == BackendTypes.DATABASE:
        backend.put(chunk_size=chunk_size, ignore_errors=ignore_errors)
    elif backend_type is None:
        msg = "Cannot find an implemented backend type for backend %s"
        logger.error(msg, backend)
        raise UnsupportedBackendException(msg, backend)


@backends_options(name="list", backends=STORAGE_BACKENDS)
@click.option(
    "-n/-a",
    "--new/--all",
    default=False,
    help="List not fetched (or all) archives",
)
@click.option(
    "-D/-I",
    "--details/--ids",
    default=False,
    help="Get archives detailled output (JSON)",
)
def list_(details, new, backend, **options):
    """List available archives from a configured storage backend"""

    logger.info("Listing archives for the configured %s backend", backend)
    logger.debug("Fetch details: %s", str(details))
    logger.debug("Backend parameters: %s", options)

    storage = get_instance_from_class(
        get_class_from_name(backend, STORAGE_BACKENDS), **options
    )
    archives = storage.list(details=details, new=new)

    counter = 0
    for archive in archives:
        click.echo(json.dumps(archive) if details else archive)
        counter += 1

    if counter == 0:
        logger.warning("Configured %s backend contains no archive", backend)



"""Swift storage backend for Ralph"""

import logging
import sys
from functools import cached_property
from urllib.parse import urlparse

from swiftclient.service import SwiftService, SwiftUploadObject

from ralph.defaults import (
    SWIFT_OS_AUTH_URL,
    SWIFT_OS_IDENTITY_API_VERSION,
    SWIFT_OS_PROJECT_DOMAIN_NAME,
    SWIFT_OS_USER_DOMAIN_NAME,
)
from ralph.exceptions import BackendException, BackendParameterException
from ralph.utils import now

from ..mixins import HistoryMixin
from .base import BaseStorage

logger = logging.getLogger(__name__)


class SwiftStorage(
    HistoryMixin, BaseStorage
):  # pylint: disable=too-many-instance-attributes
    """OpenStack's Swift storage backend"""

    name = "swift"

    # pylint: disable=too-many-arguments

    def __init__(
        self,
        os_tenant_id,
        os_tenant_name,
        os_username,
        os_password,
        os_region_name,
        os_storage_url,
        os_user_domain_name=SWIFT_OS_USER_DOMAIN_NAME,
        os_project_domain_name=SWIFT_OS_PROJECT_DOMAIN_NAME,
        os_auth_url=SWIFT_OS_AUTH_URL,
        os_identity_api_version=SWIFT_OS_IDENTITY_API_VERSION,
    ):
        """Setup options for the SwiftService"""

        self.os_tenant_id = os_tenant_id
        self.os_tenant_name = os_tenant_name
        self.os_username = os_username
        self.os_password = os_password
        self.os_region_name = os_region_name
        self.os_user_domain_name = os_user_domain_name
        self.os_project_domain_name = os_project_domain_name
        self.os_auth_url = os_auth_url
        self.os_identity_api_version = os_identity_api_version
        self.container = urlparse(os_storage_url).path.rpartition("/")[-1]
        self.os_storage_url = os_storage_url.removesuffix(f"/{self.container}")

        with SwiftService(self.options) as swift:
            stats = swift.stat()
            if not stats["success"]:
                msg = "Unable to connect to the requested container: %s"
                logger.error(msg, stats["error"])
                raise BackendParameterException(msg % stats["error"])

    @cached_property
    def options(self):
        """Returns the required options for the SwiftService"""

        return {
            "os_auth_url": self.os_auth_url,
            "os_identity_api_version": self.os_identity_api_version,
            "os_password": self.os_password,
            "os_project_domain_name": self.os_project_domain_name,
            "os_region_name": self.os_region_name,
            "os_storage_url": self.os_storage_url,
            "os_tenant_id": self.os_tenant_id,
            "os_tenant_name": self.os_tenant_name,
            "os_username": self.os_username,
            "os_user_domain_name": self.os_user_domain_name,
        }

    def list(self, details=False, new=False):
        """List files in the storage backend"""

        archives_to_skip = set()
        if new:
            archives_to_skip = set(self.get_command_history(self.name, "fetch"))
        with SwiftService(self.options) as swift:
            for page in swift.list(self.container):
                if not page["success"]:
                    msg = "Failed to list container %s: %s"
                    logger.error(msg, page["container"], page["error"])
                    raise BackendException(msg % (page["container"], page["error"]))
                for archive in page["listing"]:
                    if new and archive["name"] in archives_to_skip:
                        continue
                    yield archive if details else archive["name"]

    def url(self, name):
        """Get `name` file absolute URL"""

        # What's the purpose of this function ? Seems not used anywhere.
        return f"{self.options.get('os_storage_url')}/{name}"

    def read(self, name, chunk_size=None):
        """Read `name` object and stream its content in chunks of (max) 2 ** 16

        Why chunks of (max) 2 ** 16 ?
            Because SwiftService opens a file to stream the object into:
            See swiftclient.service.py:2082 open(filename, 'rb', DISK_BUFFER)
            Where filename = "/dev/stdout" and DISK_BUFFER = 2 ** 16
        """

        logger.debug("Getting archive: %s", name)

        with SwiftService(self.options) as swift:
            download = next(swift.download(self.container, [name], {"out_file": "-"}))
        if "contents" not in download:
            msg = "Failed to download %s: %s"
            logger.error(msg, download["object"], download["error"])
            raise BackendException(msg % (download["object"], download["error"]))
        size = 0
        for chunk in download["contents"]:
            logger.debug("Chunk %s", len(chunk))
            size += len(chunk)
            sys.stdout.buffer.write(chunk)

        # Archive fetched, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "command": "fetch",
                "id": name,
                "size": size,
                "fetched_at": now(),
            }
        )

    def write(self, name, chunk_size=None, overwrite=False):
        """Write content to the `name` target in chunks of (max) 2 ** 16"""

        if not overwrite and name in list(self.list()):
            msg = "%s already exists and overwrite is not allowed"
            logger.error(msg, name)
            raise FileExistsError(msg % name)

        logger.debug("Creating archive: %s", name)

        swift_object = SwiftUploadObject(sys.stdin.buffer, object_name=name)
        with SwiftService(self.options) as swift:
            for upload in swift.upload(self.container, [swift_object]):
                if not upload["success"]:
                    logger.error(upload["error"])
                    raise BackendException(upload["error"])

        # Archive written, add a new entry to the history
        self.append_to_history(
            {
                "backend": self.name,
                "command": "push",
                "id": name,
                "pushed_at": now(),
            }
        )



"""
Ralph tracking logs filters.
"""
from .exceptions import EventKeyError


def anonymous(events):
    """Remove anonymous events.

    Args:
        events (DataFrame): events to filter

    Returns:
        Filtered pandas DataFrame.

    """

    if events.get("username", None) is None:
        raise EventKeyError(
            "Cannot filter anonymous filters without 'username' column."
        )
    return events.loc[lambda df: df["username"] != "", :]



"""Elasticsearch storage backend for Ralph"""

import json
import logging
import sys

from elasticsearch import Elasticsearch
from elasticsearch.helpers import scan, streaming_bulk

from .base import BaseDatabase

logger = logging.getLogger(__name__)


class ESDatabase(BaseDatabase):
    """Elasticsearch database backend"""

    name = "es"

    def __init__(self, hosts: list, index: str, client_options: dict = None):
        """Instantiate the Elasticsearch client.

        hosts: supposed to be a list
        index: the index name
        es_client_options: a dict of valid options for the Elasticsearch class
                           initialization

        """
        if client_options is None:
            client_options = {}

        self._hosts = hosts
        self.index = index
        self.client = Elasticsearch(self._hosts, **client_options)

    def get(self, chunk_size=500):
        """Get index documents and stream them"""

        for document in scan(self.client, index=self.index, size=chunk_size):
            sys.stdout.buffer.write(
                bytes(json.dumps(document.get("_source")) + "\n", encoding="utf-8")
            )

    def to_documents(self, stream, get_id):
        """Convert stream lines to ES documents"""

        for line in stream:
            item = json.loads(line)
            yield {
                "_index": self.index,
                "_id": get_id(item),
                "_source": item,
            }

    def put(self, chunk_size=500, ignore_errors=False):
        """Write documents streamed from the standard input to the instance index"""

        logger.debug(
            "Start writing to the %s index (chunk size: %d)", self.index, chunk_size
        )

        documents = 0
        for success, action in streaming_bulk(
            client=self.client,
            actions=self.to_documents(sys.stdin, lambda d: d.get("id", None)),
            chunk_size=chunk_size,
            raise_on_error=(not ignore_errors),
        ):
            documents += success
            logger.debug(
                "Wrote %d documents [action: %s ok: %d]", documents, action, success
            )



"""
Ralph exceptions.
"""


class BackendException(Exception):
    """Raised when a backend has a failure"""


class BackendParameterException(Exception):
    """Raised when a backend parameter value is not valid"""


class ConfigurationException(Exception):
    """Raised when the configuration is not valid"""


class EventKeyError(Exception):
    """Raised when an expected event key has not been found."""


class UnsupportedBackendException(Exception):
    """Raised when trying to use an unsupported backend type"""



"""Backend mixins for Ralph"""

import json
import logging

from ralph.defaults import HISTORY_FILE

logger = logging.getLogger(__name__)


class HistoryMixin:
    """Handle backend download history to avoid fetching same files multiple
    times if they are already available."""

    @property
    def history(self):
        """Get backend history"""

        logging.debug("Loading history file: %s", str(HISTORY_FILE))

        if not hasattr(self, "_history"):
            try:
                with HISTORY_FILE.open() as history_file:
                    self._history = json.load(history_file)
            except FileNotFoundError:
                self._history = []
        return self._history

    # pylint: disable=no-self-use
    def write_history(self, history):
        """Write given history as a JSON file"""

        logging.debug("Writing history file: %s", str(HISTORY_FILE))

        if not HISTORY_FILE.parent.exists():
            HISTORY_FILE.parent.mkdir(parents=True)

        with HISTORY_FILE.open("w") as history_file:
            json.dump(history, history_file)

        # Update history
        self._history = history

    def clean_history(self, selector):
        """Clean selected events from the history.

        selector: a callable that selects events that need to be removed
        """
        self._history = list(filter(lambda event: not selector(event), self.history))
        self.write_history(self._history)

    def append_to_history(self, event):
        """Append event to history"""

        self.write_history(self.history + [event])

    def get_command_history(self, backend_name, command):
        """Returns a set of entry ids from the history for a command and backend_name"""

        return [
            entry["id"]
            for entry in filter(
                lambda e: e["backend"] == backend_name and e["command"] == command,
                self.history,
            )
        ]



"""Default configurations for Ralph"""

from enum import Enum
from os import environ
from pathlib import Path

import yaml
from click import get_app_dir

from ralph.exceptions import ConfigurationException


class DatabaseBackends(Enum):
    """Enumerate active database backend modules.

    Adding an entry to this enum will make it available to the CLI.
    """

    ES = "ralph.backends.database.es.ESDatabase"


class Parsers(Enum):
    """Enumerate active parsers modules.

    Adding an entry to this enum will make it available to the CLI.
    """

    GELF = "ralph.parsers.GELFParser"


class StorageBackends(Enum):
    """Enumerate active storage backend modules.

    Adding an entry to this enum will make it available to the CLI.
    """

    LDP = "ralph.backends.storage.ldp.LDPStorage"
    FS = "ralph.backends.storage.fs.FSStorage"
    SWIFT = "ralph.backends.storage.swift.SwiftStorage"


def load_config(config_file_path):
    """Return a dictionnary representing Ralph's configuration"""

    try:
        with open(config_file_path) as config_file:
            return yaml.safe_load(config_file)
    except yaml.scanner.ScannerError as exc:
        raise ConfigurationException("Configuration could not be loaded") from exc
    except FileNotFoundError:
        return None


def config(key, default_value):
    """
    Get a value based on its key returning the first of (in order):
        1. Environment
        2. Config file
        3. default_value
    """

    value = environ.get(key, None)
    if value is not None:
        return value

    if CONFIG is not None and key in CONFIG:
        return CONFIG[key]

    return default_value


DEFAULT_LOGGING_CONFIG = {
    "version": 1,
    "propagate": True,
    "formatters": {
        "ralph": {"format": "%(asctime)-23s %(levelname)-8s %(name)-8s %(message)s"},
    },
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": "INFO",
            "stream": "ext://sys.stderr",
            "formatter": "ralph",
        },
    },
    "loggers": {
        "ralph": {
            "handlers": ["console"],
            "level": "INFO",
        },
        "swiftclient": {
            "handlers": ["console"],
            "level": "ERROR",
        },
    },
}

APP_DIR = Path(environ.get("RALPH_APP_DIR", get_app_dir("ralph")))
CONFIG_FILE = APP_DIR / "config.yml"
CONFIG = load_config(CONFIG_FILE)
ENVVAR_PREFIX = "RALPH"
DEFAULT_GELF_PARSER_CHUNCK_SIZE = config("RALPH_DEFAULT_GELF_PARSER_CHUNCK_SIZE", 5000)
DEFAULT_BACKEND_CHUNCK_SIZE = config("RALPH_DEFAULT_BACKEND_CHUNCK_SIZE", 500)
FS_STORAGE_DEFAULT_PATH = Path(
    config("RALPH_FS_STORAGE_DEFAULT_PATH", APP_DIR / "archives")
)
HISTORY_FILE = Path(config("RALPH_HISTORY_FILE", APP_DIR / "history.json"))
LOGGING_CONFIG = config("RALPH_LOGGING", DEFAULT_LOGGING_CONFIG)
SENTRY_DSN = config("RALPH_SENTRY_DSN", None)
SWIFT_OS_AUTH_URL = config("RALPH_SWIFT_OS_AUTH_URL", "https://auth.cloud.ovh.net/")
SWIFT_OS_IDENTITY_API_VERSION = config("RALPH_SWIFT_OS_IDENTITY_API_VERSION", "3")
SWIFT_OS_PROJECT_DOMAIN_NAME = config("RALPH_SWIFT_OS_PROJECT_DOMAIN_NAME", "Default")
SWIFT_OS_USER_DOMAIN_NAME = config("RALPH_SWIFT_OS_USER_DOMAIN_NAME", "Default")
EXECUTION_ENVIRONMENT = config("RALPH_EXECUTION_ENVIRONMENT", "development")



"""Utilities for Ralph"""

import datetime
import logging
from importlib import import_module

from ralph.backends import BackendTypes
from ralph.backends.database.base import BaseDatabase as BaseDatabaseBackend
from ralph.backends.storage.base import BaseStorage as BaseStorageBackend


# Taken from Django utilities
# https://docs.djangoproject.com/en/3.1/_modules/django/utils/module_loading/#import_string
def import_string(dotted_path):
    """
    Import a dotted module path and return the attribute/class designated by the
    last name in the path. Raise ImportError if the import failed.
    """
    try:
        module_path, class_name = dotted_path.rsplit(".", 1)
    except ValueError as err:
        raise ImportError("%s doesn't look like a module path" % dotted_path) from err

    module = import_module(module_path)

    try:
        return getattr(module, class_name)
    except AttributeError as err:
        raise ImportError(
            'Module "%s" does not define a "%s" attribute/class'
            % (module_path, class_name)
        ) from err


def get_backend_type(backend_class):
    """Get backend type from a backend class"""

    if BaseStorageBackend in backend_class.__mro__:
        return BackendTypes.STORAGE
    if BaseDatabaseBackend in backend_class.__mro__:
        return BackendTypes.DATABASE
    return None


def get_class_names(modules):
    """Get class name attributes from class modules list"""

    return [import_string(module).name for module in modules]


def get_class_from_name(name, modules):
    """Get class given its name in a modules list"""

    for module in modules:
        klass = import_string(module)
        if klass.name == name:
            return klass
    raise ImportError(f"{name} class is not available")


def get_instance_from_class(klass, **init_parameters):
    """Return a class instance given class-name-prefixed init parameters"""

    # Filter backend-related parameters. Parameter name is supposed to start
    # with the backend name
    names = filter(lambda p: p.startswith(klass.name), init_parameters.keys())
    parameters = {
        name.replace(f"{klass.name}_", ""): init_parameters[name] for name in names
    }

    return klass(**parameters)


def get_root_logger():
    """Get main Ralph logger"""

    ralph_logger = logging.getLogger("ralph")
    ralph_logger.propagate = True

    return ralph_logger


def now():
    """Return the current UTC time in ISO format"""

    return datetime.datetime.now(tz=datetime.timezone.utc).isoformat()



"""OVH's LDP storage backend for Ralph"""

import logging
import sys

import ovh
import requests

from ralph.exceptions import BackendParameterException
from ralph.utils import now

from ..mixins import HistoryMixin
from .base import BaseStorage

logger = logging.getLogger(__name__)


class LDPStorage(HistoryMixin, BaseStorage):
    """OVH's LDP storage backend"""

    # pylint: disable=too-many-arguments

    name = "ldp"

    def __init__(
        self,
        endpoint,
        application_key,
        application_secret,
        consumer_key,
        service_name=None,
        stream_id=None,
    ):
        self._endpoint = endpoint
        self._application_key = application_key
        self._application_secret = application_secret
        self._consumer_key = consumer_key
        self.service_name = service_name
        self.stream_id = stream_id

        self.client = ovh.Client(
            endpoint=self._endpoint,
            application_key=self._application_key,
            application_secret=self._application_secret,
            consumer_key=self._consumer_key,
        )

    @property
    def _archive_endpoint(self):
        if None in (self.service_name, self.stream_id):
            msg = "LDPStorage backend instance requires to set both service_name and stream_id"
            logger.error(msg)
            raise BackendParameterException(msg)
        return f"/dbaas/logs/{self.service_name}/output/graylog/stream/{self.stream_id}/archive"

    def _details(self, name):
        """Get name archive details

        Expected JSON response looks like:

            {
                "archiveId": "5d49d1b3-a3eb-498c-9039-6a482166f888",
                "createdAt": "2020-06-18T04:38:59.436634+02:00",
                "filename": "2020-06-16.gz",
                "md5": "01585b394be0495e38dbb60b20cb40a9",
                "retrievalDelay": 0,
                "retrievalState": "sealed",
                "sha256": "645d8e21e6fdb8aa7ffc507acf091ada39dbdc9ce612d06df8dcf67cb29a45ca",
                "size": 67906662,
            }
        """
        return self.client.get(f"{self._archive_endpoint}/{name}")

    def url(self, name):
        """Get archive absolute URL"""

        download_url_endpoint = f"{self._archive_endpoint}/{name}/url"

        response = self.client.post(download_url_endpoint)
        download_url = response.get("url")
        logger.debug("Temporary URL: %s", download_url)

        return download_url

    def list(self, details=False, new=False):
        """List archives for a given stream.

        details: get detailled information about archives instead of their ids

        new: given the history, list only not already fetched archives

        """

        list_archives_endpoint = self._archive_endpoint
        logger.debug("List archives endpoint: %s", list_archives_endpoint)
        logger.debug("List archives details: %s", str(details))

        archives = self.client.get(list_archives_endpoint)
        logger.debug("Found %d archives", len(archives))

        if new:
            archives = set(archives) - set(self.get_command_history(self.name, "fetch"))
            logger.debug("New archives: %d", len(archives))

        for archive in archives:
            yield self._details(archive) if details else archive

    def read(self, name, chunk_size=4096):
        """Read the `name` archive file and stream its content"""

        logger.debug("Getting archive: %s", name)

        # Get detailled information about the archive to fetch
        details = self._details(name)

        # Stream response (archive content)
        with requests.get(self.url(name), stream=True) as result:
            result.raise_for_status()
            for chunk in result.iter_content(chunk_size=chunk_size):
                sys.stdout.buffer.write(chunk)

        # Archive is supposed to have been fully fetched, add a new entry to
        # the history.
        self.append_to_history(
            {
                "backend": self.name,
                "command": "fetch",
                "id": name,
                "filename": details.get("filename"),
                "size": details.get("size"),
                "fetched_at": now(),
            }
        )

    def write(self, name, chunk_size=4096, overwrite=False):
        """LDP storage backend is read-only, calling this method will raise an error"""

        msg = "LDP storage backend is read-only, cannot write to %s"
        logger.error(msg, name)
        raise NotImplementedError(msg % name)



#!/usr/bin/env python

"""
Ralph main entrypoint.

This script is a POC for now, we aim to make ralph a nice CLI to use.
"""

import sentry_sdk

from ralph.defaults import EXECUTION_ENVIRONMENT, SENTRY_DSN

from . import __version__, cli

if SENTRY_DSN is not None:
    sentry_sdk.init(
        dsn=SENTRY_DSN,
        traces_sample_rate=1.0,
        release=__version__,
        environment=EXECUTION_ENVIRONMENT,
        max_breadcrumbs=50,
    )

cli.cli()



""""Ralph logging manager"""

from logging.config import dictConfig

from ralph.defaults import LOGGING_CONFIG
from ralph.exceptions import ConfigurationException


def configure_logging():
    """Set up Ralph logging configuration"""

    try:
        dictConfig(LOGGING_CONFIG)
    except Exception as error:
        raise ConfigurationException("Improperly configured logging") from error



"""
Ralph tracking logs parsers.
"""

import logging
from abc import ABC, abstractmethod
from pathlib import Path

import pandas as pd

from .defaults import DEFAULT_GELF_PARSER_CHUNCK_SIZE

logger = logging.getLogger(__name__)


class BaseParser(ABC):
    """Base tracking logs parser."""

    name = "base"

    @abstractmethod
    def parse(self, input_file, chunksize=1):
        """Parse GELF formatted logs (one json string event per row).

        Args:
            input_file (string): Path to the log file to parse.
            chunksize (int): The amount of log records to process at a time.

        Yields:
            event: raw event as extracted from its container

        """


class GELFParser(BaseParser):
    """GELF formatted logs parser.

    Documentation of the GELF log format can be found in the Graylog project's
    documentation: https://docs.graylog.org/en/latest/pages/gelf.html
    """

    name = "gelf"

    def parse(self, input_file, chunksize=DEFAULT_GELF_PARSER_CHUNCK_SIZE):
        """Parse GELF formatted logs (one json string event per row).

        Args:
            input_file (string): Path to the log file to parse (could be
                gunzipped).
            chunksize (int): The amount of log records to process at a time. A
                value between 3.000 and 10.000 seems to be a reasonnable choice
                to parse 1.5M records in a few minutes (typically 2 or 3
                minutes on a modern computer).

        Yields:
            event: events raw short_message string

        """
        logger.info("Parsing: %s", input_file)

        if isinstance(input_file, str) and not Path(input_file).exists():
            msg = "Input GELF log file '%s' does not exist"
            logger.error(msg, input_file)
            raise OSError(msg % (input_file))

        chunks = pd.read_json(input_file, lines=True, chunksize=chunksize)
        for chunk in chunks:
            for event in chunk["short_message"].values:
                yield event



"""Base storage backend for Ralph"""

from abc import ABC, abstractmethod


class BaseDatabase(ABC):
    """Base database backend interface"""

    name = "base"

    @abstractmethod
    def get(self, chunk_size=10):
        """Read chunk_size records and stream them to stdout"""

    @abstractmethod
    def put(self, chunk_size=10, ignore_errors=False):
        """Write chunk_size records from stdin"""
